-- vim: ts=4 sw=4 expandtab

require "Apollo"
require "ApolloTimer"
require "XmlDoc"
require "GameLib"
require "ChallengesLib"

------------------------------------
-- Boilerplate
------------------------------------

local Addon = {}
 
function Addon:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self

    self.classById = {}
    for k, v in pairs(GameLib.CodeEnumClass) do
        self.classById[v] = k
    end

    return o
end

function Addon:Init()
    Apollo.RegisterAddon(self, false, "", {})
end

------------------------------------
-- Common
------------------------------------

local ti = table.insert
local function toboolean(val) if val then return true else return false end end

function Addon:OnLoad()
    self.nameplates = {}
    self.scanResults = {}

    self:resetSettings()

    Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
    Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)

    Apollo.RegisterEventHandler("UnitCreated", "RadarOnUnitCreated", self)
    Apollo.RegisterEventHandler("UnitDestroyed", "RadarOnUnitDestroyed", self)

    Apollo.RegisterEventHandler("PlayerChanged", "OnPlayerChanged", self)

    self.xmlDoc = XmlDoc.CreateFromFile("ClearNameplates.xml")
    self.xmlDoc:RegisterCallback("OnDocumentReady", self)
end

function Addon:OnDocumentReady()
    self.me = GameLib.GetPlayerUnit() -- like CRB Nameplates

    self:wpInit()

    self.npTimer = ApolloTimer.Create(0.25, true, "UpdateNameplates", self)

    self.radarWindow = Apollo.LoadForm("ClearNameplates.xml", "ScanResultsWindow", nil, self)
    self.radarWindow:Show(self.radarEnabled)

    self.radarTimer = ApolloTimer.Create(0.5, true, "RadarScan", self)

    Apollo.RegisterSlashCommand("cnp", "OnCnpCommand", self)
end

function Addon:OnPlayerChanged()
    self.me = GameLib.GetPlayerUnit()
end

function Addon.distanceToUnit(unit, me)
    local unitPos = unit:GetPosition()
    local selfPos = me:GetPosition()

    if selfPos == nil or unitPos == nil then return 0 end -- in taxi

    return (math.abs(selfPos.x - unitPos.x)) + (math.abs(selfPos.y - unitPos.y)) + (math.abs(selfPos.z - unitPos.z))
end

------------------------------------
-- The Logic of Nameplates
------------------------------------

function Addon:OnUnitCreated(unit)
    local unitId = unit:GetId()
    local npInfo = self.nameplates[unitId]

    if npInfo then
--        Addon.log("unit> already exists, skipping", unit)
        return
    end

    local unitType = unit:GetType()

    if self:shouldHideNameplate(unit) then return end

    self.nameplates[unit:GetId()] = {
        unit = unit,
        window = nil
    }

--    Addon.log("unit>", unit)

end

function Addon:OnUnitDestroyed(unit)
    local unitId = unit:GetId()
--    Addon.log(string.format("<unit %s %s", unitId, unit:GetName()))
    self:removeNameplate(unit)

    -- My character got destroyed
    if self.me and self.me:GetId() == unitId then
        self.me = nil
        return
    end

end

function Addon:removeNameplate(unit)
    local unitId = unit:GetId()
    local npInfo = self.nameplates[unitId]
    if not npInfo then return end

    self.nameplates[unitId] = nil

    if npInfo.window then
        self:wpReturn(npInfo.window)
        npInfo.window = nil
    end
end

function Addon:shouldHideNameplate(unit)
    return (not unit:ShouldShowNamePlate()) and (unit:GetType() ~= "Harvest")
end

function Addon:UpdateNameplates()
    if not self.me then return end

    local window, unit

    local showByKeyDown = Apollo.IsControlKeyDown()
    local activeChallenges = ChallengesLib.GetActiveChallengeList()

    local removeUnits = {}

    for unitId, npInfo in pairs(self.nameplates) do

        repeat

            unit = npInfo.unit
            window = npInfo.window

--            Addon.log("Iterating", unit)

            local unitOccluded = unit:IsOccluded()
            local windowOccluded = (window and window:IsOccluded()) or false

            if not unit:IsValid() or (window and not window:IsValid()) then
                ti(removeUnits, unit)
--                Addon.log("Invalid window/unit", unit)
                break
            end

            local myself = self.me:GetId() == unitId
            local player = unit:GetType() == "Player"
            local npc = unit:GetType() == "NonPlayer"
            local pvp = unit:IsPvpFlagged()
            local distance = Addon.distanceToUnit(unit, self.me)
            local mounted = unit:IsMounted()

            if window then
                if windowOccluded then -- WHAT DOES THAT MEAN? OMG IT'S KILLING ME.
                    window:Show(false, true)
                    break
                end
                if self:shouldHideNameplate(unit) then
                    window:Show(false, true)
                    break
                end
                if not mounted and unitOccluded then
                    window:Show(false, true)
                    break
                end
                if mounted and unitOccluded and unit:GetUnitMount():IsOccluded() then
                    window:Show(false, true)
                    break
                end
            end

            local disposition = unit:GetDispositionTo(self.me)
            local hostile = disposition == Unit.CodeEnumDisposition.Hostile
            local closeToMe = distance <= 130

            local activationStates = unit:GetActivationState()
            local rewardInfos = unit:GetRewardInfo()
            local questUnit = not player and Addon.isQuestUnit(activationStates, rewardInfos)
            local challengeUnit = not player and Addon.isChallengeUnit(rewardInfos, activeChallenges)

            local otherHostilePlayer = not myself and player and pvp
            local closeHostile = hostile and closeToMe
            local interestingUnit = not player and (questUnit or Addon.isInterestingUnit(activationStates))
            local fightingMe = self.me:IsInCombat() and unit:IsTaggedByMe()
            local targeted = self.me:GetTarget() and self.me:GetTarget():GetId() == unitId
            local dead = unit:IsDead()

            local weirdNpc = unit:GetType() == "Simple"

            local zoneFactionToPlayerFaction = {
                [GameLib.CodeEnumZonePvpRules.ExileStronghold] = Unit.CodeEnumFaction.ExilesPlayer,
                [GameLib.CodeEnumZonePvpRules.DominionStronghold] = Unit.CodeEnumFaction.DominionPlayer
            }

            local zoneFaction = GameLib.GetCurrentZonePvpRules()
            local insideEnemyTerritory = zoneFactionToPlayerFaction[zoneFaction] and zoneFactionToPlayerFaction[zoneFaction] ~= self.me:GetFaction()
            local noncombatantHostile = insideEnemyTerritory and unit:IsSoftKill() and npc

            if (targeted or otherHostilePlayer or closeHostile or interestingUnit or fightingMe or showByKeyDown)
                and not (dead and npc) and not weirdNpc and not noncombatantHostile or showByKeyDown then

                if window == nil then
                    window = self:wpBorrow()
                    window:SetUnit(unit, 1)
                    npInfo.window = window
--                    Addon.log(string.format("window> %s", window:GetId()), unit)
                    break
                end

                local content

                if self.altUnitText then
                    content = window:FindChild("Content")
                    if content then
                        content:SetText(
                            self:calcAltUnitText(unit, npc, player, questUnit, challengeUnit, distance, pvp, disposition)
                        )
                    end
                end

                if not content then
                    window:SetText(self:calcUnitText(unit, npc, player, hostile, questUnit, challengeUnit))
                    window:SetTextColor(self:calcUnitColor(npc, player, pvp, disposition, distance))
                end

                window:Show(true, true)

            elseif window then

                window:Show(false, true)

            end

        until true

    end

    for _, unit in pairs(removeUnits) do
--        Addon.log("<unit-manual", unit)
        self:removeNameplate(unit)
    end

end

function Addon.isQuestUnit(activationState, rewardInfo)
    local res = toboolean(
        activationState.QuestReward
                or activationState.QuestNew
                or activationState.QuestNewMain
                or activationState.QuestReceiving
                or activationState.TalkTo
    )
    if res then return res end
    if rewardInfo then
        for _, reward in pairs(rewardInfo) do
            if reward.strType == "Quest" then return true end
        end
    end
    return false
end

function Addon.isChallengeUnit(rewardInfos, activeChallenges)
    if not rewardInfos then return false end

    local info
    for _, infoEl in pairs(rewardInfos) do
        if infoEl.strType == "Challenge" then
            info = infoEl
            break
        end
    end
    if not info then return false end

    local challenge
    for _, challengeEl in pairs(activeChallenges) do
        if challengeEl:GetId() == info.idChallenge then
            challenge = challengeEl
            break
        end
    end
    if not challenge then return false end

    return challenge:IsActivated() and not challenge:IsInCooldown() and not challenge:ShouldCollectReward()
end

function Addon.isInterestingUnit(activationState)
    return toboolean(
        activationState.Vendor
                or activationState.FlightPath
                or activationState.FlightPathNew
                or activationState.FlightPathSettler
    )
end

function Addon:calcHealthBar(text, unit, hp)
    local sp = unit:GetShieldCapacity()
    local spMax = sp and math.max(sp, unit:GetShieldCapacityMax())
    local ap = unit:GetAbsorptionValue()

    if ap and ap > 0 then ti(text, "ӿ") end

    local healthDots = math.ceil(hp / (unit:GetMaxHealth() / 5.0))
    local shieldDashes = math.max(0, math.ceil(sp / (spMax / 4.0)))
    for bulletIdx=1,5 do
        ti(text, ((bulletIdx <= healthDots) and "●") or "○")
        if bulletIdx <= shieldDashes then
            ti(text, "\205\139") -- unicode combiner mark
        end
    end

    if ap and ap > 0 then ti(text, "ӿ") end

    ti(text, "\n")
end

function Addon:calcUnitName(text, unit, npc, player, hostile, quest, challenge)
    if player and hostile then
        ti(text, self.classById[unit:GetClassId()])
        return table.concat(text)
    end

    local unitName = unit:GetName()
    local difficulty = (unit:GetDifficulty() > 2 and string.format(" /%d", unit:GetDifficulty())) or nil

    if player then
        ti(text, unitName)
        return table.concat(text)
    end

    if quest then ti(text, "◦") end
    if challenge then ti(text, "©") end
    ti(text, " ")
    ti(text, unitName)
    if npc and hostile and difficulty then ti(text, difficulty) end
end

function Addon:calcUnitText(unit, npc, player, hostile, quest, challenge)

--    if true then
--        return string.format("ӿ●%s●%s●%s○%s○ӿ\n◦ © A Scary Creature",
--            "\205\139", "\205\139", "\205\139", "\205\139"
--        )
--    end

    local text = { "" }

    local hp = unit:GetHealth() or 0
    if hp > 0 and not (npc and not hostile) then
        self:calcHealthBar(text, unit, hp)
    end

    self:calcUnitName(text, unit, npc, player, hostile, quest, challenge)

    return table.concat(text)
end

function Addon:calcUnitAlpha(distance)
    local alphaNum
    if distance < 45 then alphaNum = 255
    else alphaNum = math.max(128, 255 - math.floor(((distance - 35) * 1.25)))
    end
    return alphaNum
end

function Addon:calcUnitColor(npc, player, pvp, disposition, distance)
    local hostile = disposition == Unit.CodeEnumDisposition.Hostile
    local friendly = disposition == Unit.CodeEnumDisposition.Friendly

    local rgb
    repeat
        if hostile and player and pvp then rgb = self.hostilePlayerPvpColor; break; end
        if hostile and player then rgb = self.hostilePlayerNotpvpColor; break; end
        if hostile then rgb = self.hostileColor; break; end
        if friendly then rgb = self.friendlyColor; break; end
        rgb = self.neutralColor
    until true

    return string.format("%02x%s", self:calcUnitAlpha(distance), rgb)
end

function Addon:createNpWindow()
    local win

    if self.altUnitText then
        win = Apollo.LoadForm(self.xmlDoc, "AltNameplateWindow", "InWorldHudStratum", self)
    else
        win = Apollo.LoadForm(self.xmlDoc, "NameplateWindow", "InWorldHudStratum", self)
        win:SetFont(string.format("CRB_Interface%d_BBO", self.fontSize))
    end
    return win
end

function Addon:destroyNpWindows()
    for _,npInfo in pairs(self.nameplates) do
        if not npInfo.window then break end
        npInfo.window:Show(false, true)
        npInfo.window:Destroy()
        npInfo.window = nil
    end
end

------------------------------------
-- Boilerplate
------------------------------------

local ClearNameplatesInst = Addon:new()
ClearNameplatesInst:Init()
